#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import argparse
import asyncio
import os

import sys
import aiohttp

import octobot_commons.logging as logging

import octobot_tentacles_manager.api as api
import octobot_tentacles_manager.api.updater as updater_api
import octobot_tentacles_manager.api.installer as installer_api
import octobot_tentacles_manager.api.uninstaller as uninstaller_api
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager


async def _handle_package_manager_command(starting_args,
                                          tentacles_urls,
                                          target_dir,
                                          bot_install_dir,
                                          single_tentacle_path,
                                          single_tentacle_type,
                                          export_tentacles_output,
                                          exported_tentacles_package,
                                          packed_tentacles_output,
                                          quite_mode,
                                          cythonize) -> int:
    error_count = 0
    LOGGER = logging.get_logger(f"{octobot_tentacles_manager.PROJECT_NAME}-CLI")
    async with aiohttp.ClientSession() as aiohttp_session:
        include_dev_mode = starting_args.include_dev_mode
        include_tentacles_export = starting_args.include_tentacles_export
        if starting_args.creator:
            error_count = api.start_tentacle_creator({}, starting_args.creator)
        elif starting_args.repair:
            error_count = await api.repair_installation(bot_path=target_dir)
        elif starting_args.export:
            error_count = await api.create_tentacles_package(export_tentacles_output,
                                                             target_dir,
                                                             exported_tentacles_package=exported_tentacles_package,
                                                             in_zip=False,
                                                             with_dev_mode=include_dev_mode,
                                                             upload_details=starting_args.upload_package_export,
                                                             cythonize=cythonize)
            if include_tentacles_export:
                error_count += await api.create_all_tentacles_bundle(
                    output_dir=export_tentacles_output,
                    tentacles_folder=target_dir,
                    exported_tentacles_package=exported_tentacles_package,
                    in_zip=False,
                    should_remove_artifacts_after_use=True,
                    upload_url=starting_args.upload_tentacles_export[0]
                    if len(starting_args.upload_tentacles_export) > 0 else None,
                    with_dev_mode=include_dev_mode,
                    cythonize=cythonize)
        elif starting_args.pack:
            error_count = await api.create_tentacles_package(packed_tentacles_output,
                                                             target_dir,
                                                             exported_tentacles_package=exported_tentacles_package,
                                                             in_zip=True,
                                                             with_dev_mode=include_dev_mode,
                                                             upload_details=starting_args.upload_package_export,
                                                             cythonize=cythonize)
            if include_tentacles_export:
                error_count += await api.create_all_tentacles_bundle(
                    output_dir=export_tentacles_output,
                    tentacles_folder=target_dir,
                    exported_tentacles_package=exported_tentacles_package,
                    in_zip=True,
                    with_dev_mode=include_dev_mode,
                    should_remove_artifacts_after_use=True,
                    upload_url=starting_args.upload_tentacles_export[0]
                    if len(starting_args.upload_tentacles_export) > 0 else None,
                    cythonize=cythonize)
        elif single_tentacle_path:
            error_count = await api.install_single_tentacle(single_tentacle_path,
                                                            single_tentacle_type,
                                                            bot_path=target_dir,
                                                            aiohttp_session=aiohttp_session,
                                                            bot_install_dir=bot_install_dir)
        elif starting_args.nexus_upload:
            if 2 <= len(starting_args.nexus_upload) <= 3:
                file_or_folder_path = starting_args.nexus_upload[1]
                file_or_folder_alias = starting_args.nexus_upload[2] if len(starting_args.nexus_upload) == 3 else None
                error_count += await api.upload_file_or_folder_to_nexus(nexus_path=starting_args.nexus_upload[0],
                                                                        artifact_path=file_or_folder_path,
                                                                        artifact_alias=file_or_folder_alias)
            else:
                raise Exception("Missing arguments. Usage : <nexus_path> <file_or_folder_path> [file_or_folder_alias].")

        elif not (starting_args.all or starting_args.tentacle_names):
            LOGGER.error("Please provide at least one tentacle name or add the '--all' parameter")
            return 1
        elif starting_args.install:
            for tentacles_url in tentacles_urls:
                if tentacles_url is None:
                    LOGGER.error("Please provide a tentacle path or URL")
                    error_count += 1
                elif starting_args.all:
                    error_count += await api.install_all_tentacles(tentacles_url,
                                                                   bot_path=target_dir,
                                                                   aiohttp_session=aiohttp_session,
                                                                   quite_mode=quite_mode,
                                                                   bot_install_dir=bot_install_dir)
                else:
                    error_count += await api.install_tentacles(starting_args.tentacle_names,
                                                               tentacles_url,
                                                               bot_path=target_dir,
                                                               aiohttp_session=aiohttp_session,
                                                               quite_mode=quite_mode,
                                                               bot_install_dir=bot_install_dir)
        elif starting_args.update:
            for tentacles_url in tentacles_urls:
                if tentacles_url is None:
                    LOGGER.error("Please provide a tentacle path or URL")
                    error_count += 1
                elif starting_args.all:
                    error_count += await api.update_all_tentacles(tentacles_url,
                                                                  bot_path=target_dir,
                                                                  aiohttp_session=aiohttp_session,
                                                                  quite_mode=quite_mode)
                else:
                    error_count += await api.update_tentacles(starting_args.tentacle_names,
                                                              tentacles_url,
                                                              bot_path=target_dir,
                                                              aiohttp_session=aiohttp_session,
                                                              quite_mode=quite_mode)
        elif starting_args.uninstall:
            if starting_args.all:
                error_count += await api.uninstall_all_tentacles(bot_path=target_dir,
                                                                 use_confirm_prompt=starting_args.force,
                                                                 quite_mode=quite_mode)
            else:
                error_count += await api.uninstall_tentacles(starting_args.tentacle_names,
                                                             bot_path=target_dir,
                                                             use_confirm_prompt=starting_args.force,
                                                             quite_mode=quite_mode)

    if error_count > 0:
        LOGGER.error(f"{error_count} errors occurred while processing tentacles.")
        return 1
    return 0


def handle_tentacles_manager_command(starting_args,
                                     tentacles_urls=None,
                                     single_tentacle_path=None,
                                     single_tentacle_type=None,
                                     export_tentacles_output=None,
                                     exported_tentacles_package=None,
                                     packed_tentacles_output=None,
                                     bot_install_dir=constants.DEFAULT_BOT_INSTALL_DIR,
                                     quite=False) -> int:
    tentacles_urls = [starting_args.location[0]] if starting_args.location else tentacles_urls
    target_bot_dir = starting_args.directory[0] if starting_args.directory else constants.DEFAULT_BOT_PATH
    if starting_args.single_tentacle_install:
        single_tentacle_path = starting_args.single_tentacle_install[0]
        single_tentacle_type = starting_args.single_tentacle_install[1]
    export_tentacles_output = starting_args.export[0] if starting_args.export else export_tentacles_output
    exported_tentacles_package = starting_args.export[1] if starting_args.export and len(starting_args.export) > 1 \
        else exported_tentacles_package
    packed_tentacles_output = starting_args.pack[0] if starting_args.pack else packed_tentacles_output
    quite_mode = starting_args.quite or quite
    return asyncio.run(_handle_package_manager_command(starting_args,
                                                       tentacles_urls,
                                                       target_bot_dir,
                                                       bot_install_dir,
                                                       single_tentacle_path,
                                                       single_tentacle_type,
                                                       export_tentacles_output,
                                                       exported_tentacles_package,
                                                       packed_tentacles_output,
                                                       quite_mode,
                                                       starting_args.cythonize))


def register_tentacles_manager_arguments(tentacles_parser) -> None:
    tentacles_parser.add_argument("-i", "--install", help=installer_api.USER_HELP, action='store_true')
    tentacles_parser.add_argument("-u", "--update", help=updater_api.USER_HELP, action='store_true')
    tentacles_parser.add_argument("-ui", "--uninstall", help=uninstaller_api.USER_HELP, action='store_true')
    tentacles_parser.add_argument("-r", "--repair", help="Repair a tentacles installation. "
                                                         "Fixes __init__.py files, missing main folders "
                                                         "and configuration files.", action='store_true')
    tentacles_parser.add_argument("-sti", "--single-tentacle-install", help='Single tentacle to install identified by '
                                                                            'a local path and its tentacle type.\n'
                                                                            'Example: -sti "/bot/macd_eval" '
                                                                            '"Evaluator/TA".',
                                  nargs=2)
    tentacles_parser.add_argument("-p", "--pack", help="Create a tentacle package containing all tentacles "
                                                       "in the given folder.\n"
                                                       "Specify source folder using the --directory argument.\n"
                                                       "Add --cythonize to cythonize and compile the packed tentacles."
                                                       "Example: -p myTentaclesPackage.zip -d ./tentacles", nargs=1)
    tentacles_parser.add_argument("-e", "--export", help="Export tentacles into a folder containing tentacles "
                                                         "in the given folder. Removes installation generated files. "
                                                         "Exported tentacles can be filtered to only be the ones from a"
                                                         " given package by using the package name as a second "
                                                         "argument. Specify source folder using the --directory "
                                                         "argument.\n"
                                                         "Example (all tentacles): -e myTentaclesFolder -d ./tentacles"
                                                         "\nother example (tentacles from 'OctoBot-Default-Tentacles' "
                                                         "package): -e myTentaclesFolder OctoBot-Default-Tentacles "
                                                         "-d ./tentacles", nargs="+")
    tentacles_parser.add_argument("-a", "--all", help="Apply command to all available Tentacles.", action='store_true')
    tentacles_parser.add_argument("-d", "--directory", help=f"Path to the root of the OctoBot installation folder "
                                                            f"to operate on. Default is '{constants.DEFAULT_BOT_PATH}'.",
                                  nargs=1)
    tentacles_parser.add_argument("-l", "--location", help="Tentacles package local path or url to find the "
                                                           "tentacles package to process.", nargs=1)
    tentacles_parser.add_argument("-f", "--force", help="Skip user confirmations.", action='store_true')
    tentacles_parser.add_argument("-idm", "--include-dev-mode", help="Include tentacles in dev mode in export and "
                                                                     "pack commands.", action='store_true')
    tentacles_parser.add_argument("-ite", "--include-tentacles-export", help="Include tentacles during export and "
                                                                             "pack commands.", action='store_true')
    tentacles_parser.add_argument("-ute", "--upload-tentacles-export", help="Upload tentacles during export and "
                                                                            "pack commands. Usage: <nexus_path>",
                                  nargs=1)
    tentacles_parser.add_argument("-upe", "--upload-package-export", help="Upload package during export and "
                                                                          "pack commands. "
                                                                          "Usage: <nexus_path> [package_alias]",
                                  nargs='+')
    tentacles_parser.add_argument("-c", "--creator", help="Start OctoBot Tentacles Creator.\n Examples: -c Evaluator "
                                                          "to create a new evaluator tentacles. Use: -c help to get the"
                                                          " Tentacle Creator help.", nargs='+')
    tentacles_parser.add_argument("-cy", "--cythonize", help="Option for the --pack command: cythonize and "
                                                             "compile the packed tentacles.", action='store_true')
    tentacles_parser.add_argument("-nu", "--nexus-upload", help="Upload a file or files of folder to nexus at path."
                                                                "Usage : <nexus_path> <file_or_folder_path> "
                                                                "[file_or_folder_alias]"
                                                                "Example : officials/base/ 0.4.0.zip",
                                  nargs='+')
    tentacles_parser.add_argument("-q", "--quite", help="Only display errors in logs.", action='store_true')
    tentacles_parser.add_argument("tentacle_names", nargs="*")


def main():
    parser = argparse.ArgumentParser(description=f"{octobot_tentacles_manager.PROJECT_NAME}-CLI")
    register_tentacles_manager_arguments(parser)
    args = parser.parse_args(sys.argv[1:])
    sys.exit(handle_tentacles_manager_command(args))


if __name__ == "__main__":
    main()



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_tentacles_manager.models.tentacle as tentacle_model
import octobot_tentacles_manager.models.tentacle_package as tentacle_package_model
import octobot_tentacles_manager.models.profile as profile_model

import octobot_tentacles_manager.models.metadata.tentacle_metadata as tentacle_metadata
import octobot_tentacles_manager.models.metadata.tentacle_package_metadata as tentacle_package_metadata
import octobot_tentacles_manager.models.metadata.profile_metadata as profile_metadata


class MetadataFactory:
    def __init__(self, artifact):
        self.artifact = artifact

    def create_metadata_instance(self):
        if isinstance(self.artifact, tentacle_model.Tentacle):
            return tentacle_metadata.TentacleMetadata(self.artifact)
        elif isinstance(self.artifact, tentacle_package_model.TentaclePackage):
            return tentacle_package_metadata.TentaclePackageMetadata(self.artifact)
        elif isinstance(self.artifact, profile_model.Profile):
            return profile_metadata.ProfileMetadata(self.artifact)
        return None



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from shutil import rmtree
import pytest
from copy import copy
from os.path import exists, join

from octobot_commons.constants import USER_FOLDER
from octobot_tentacles_manager.api.configurator import get_tentacles_setup_config, update_activation_configuration, \
    get_tentacles_activation
from octobot_tentacles_manager.constants import TENTACLES_PATH, DEFAULT_BOT_PATH, TENTACLES_INSTALL_TEMP_DIR
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles
from octobot_tentacles_manager.workers.install_worker import InstallWorker

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

temp_dir = "temp_tests"


async def test_update_activation_configuration():
    _cleanup(False)
    await fetch_and_extract_tentacles(temp_dir, join("tests", "static", "tentacles.zip"), None)
    worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_setup_manager.default_tentacle_config = join("tests", "static", "default_tentacle_config.json")
    assert await worker.process() == 0
    setup_config = get_tentacles_setup_config()
    default_activation = copy(get_tentacles_activation(setup_config))
    assert default_activation == {
        'Backtesting': {
            'GenericExchangeDataImporter': True
        },
        'Evaluator': {
            'InstantFluctuationsEvaluator': True,
            'OtherInstantFluctuationsEvaluator': False,
            'OverallStateAnalyser': True,
            'RedditForumEvaluator': False,
            'SecondOtherInstantFluctuationsEvaluator': False,
            'SimpleMixedStrategyEvaluator': True,
            'TextAnalysis': True
        },
        'Services': {
            'RedditService': True,
            'RedditServiceFeed': True
        },
        'Trading': {
            'DailyTradingMode': True
        }
    }
    # Did not add OtherTentacle since it is not in original activation
    assert update_activation_configuration(setup_config, {"OtherTentacle": True}, False) is False
    assert default_activation == get_tentacles_activation(setup_config)

    # No change
    assert update_activation_configuration(setup_config, {"InstantFluctuationsEvaluator": True}, False) is False
    assert default_activation == get_tentacles_activation(setup_config)

    # One change
    assert update_activation_configuration(setup_config, {"InstantFluctuationsEvaluator": False}, False) is True
    assert get_tentacles_activation(setup_config) == {
        'Backtesting': {
            'GenericExchangeDataImporter': True
        },
        'Evaluator': {
            'InstantFluctuationsEvaluator': False,
            'OtherInstantFluctuationsEvaluator': False,
            'OverallStateAnalyser': True,
            'RedditForumEvaluator': False,
            'SecondOtherInstantFluctuationsEvaluator': False,
            'SimpleMixedStrategyEvaluator': True,
            'TextAnalysis': True
        },
        'Services': {
            'RedditService': True,
            'RedditServiceFeed': True
        },
        'Trading': {
            'DailyTradingMode': True
        }
    }

    # Two changes
    assert update_activation_configuration(setup_config,
                                           {
                                              "InstantFluctuationsEvaluator": True,
                                              'RedditForumEvaluator': True
                                           },
                                           False) is True
    assert get_tentacles_activation(setup_config) == {
        'Backtesting': {
            'GenericExchangeDataImporter': True
        },
        'Evaluator': {
            'InstantFluctuationsEvaluator': True,
            'OtherInstantFluctuationsEvaluator': False,
            'OverallStateAnalyser': True,
            'RedditForumEvaluator': True,
            'SecondOtherInstantFluctuationsEvaluator': False,
            'SimpleMixedStrategyEvaluator': True,
            'TextAnalysis': True
        },
        'Services': {
            'RedditService': True,
            'RedditServiceFeed': True
        },
        'Trading': {
            'DailyTradingMode': True
        }
    }

    # Two changes with deactivate others evaluators
    assert update_activation_configuration(setup_config,
                                           {
                                              "InstantFluctuationsEvaluator": True,
                                              'RedditForumEvaluator': True
                                           },
                                           True) is False
    assert get_tentacles_activation(setup_config) == {
        'Backtesting': {
            'GenericExchangeDataImporter': True
        },
        'Evaluator': {
            'InstantFluctuationsEvaluator': True,
            'OtherInstantFluctuationsEvaluator': False,
            'OverallStateAnalyser': True,
            'RedditForumEvaluator': True,
            'SecondOtherInstantFluctuationsEvaluator': False,
            'SimpleMixedStrategyEvaluator': True,
            'TextAnalysis': True
        },
        'Services': {
            'RedditService': True,
            'RedditServiceFeed': True
        },
        'Trading': {
            'DailyTradingMode': True
        }
    }
    _cleanup()


def _tentacles_local_path():
    return join("tests", "static", "tentacles.zip")


def _cleanup(raises=True):
    if exists(TENTACLES_PATH):
        TentaclesSetupManager.delete_tentacles_arch(force=True, raises=raises, with_user_config=True)
    if exists(USER_FOLDER):
        rmtree(USER_FOLDER)
    if exists(TENTACLES_INSTALL_TEMP_DIR):
        rmtree(TENTACLES_INSTALL_TEMP_DIR)



#  Drakkar-Software OctoBot-Tentacles-Manager-Launcher
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.creators import tentacle_creator
from octobot_tentacles_manager.creators import compiled_package_manager

from octobot_tentacles_manager.creators.tentacle_creator import (
    TentacleCreator,
    CreatedTentacle,
)
from octobot_tentacles_manager.creators.compiled_package_manager import (
    cythonize_and_compile_tentacles,
)

__all__ = [
    "TentacleCreator",
    "CreatedTentacle",
    "cythonize_and_compile_tentacles",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiofiles
import zipfile
import os
import os.path as path
import shutil
import sys

import octobot_tentacles_manager.constants as constants
import octobot_commons.os_util as os_util
import octobot_commons.enums as commons_enums

DOWNLOADED_DATA_CHUNK_SIZE = 60000


async def fetch_and_extract_tentacles(tentacles_temp_dir, tentacles_path_or_url, aiohttp_session, merge_dirs=False):
    compressed_file = tentacles_path_or_url
    should_download = _is_url(tentacles_path_or_url)
    try:
        if should_download:
            if aiohttp_session is None:
                raise RuntimeError("Missing aiohttp_session argument")
            compressed_file = f"downloaded_{tentacles_temp_dir}"
            await _download_tentacles(compressed_file, tentacles_path_or_url, aiohttp_session)
        await _extract_tentacles(compressed_file, tentacles_temp_dir, merge_dirs)
    finally:
        if should_download and path.isfile(compressed_file):
            os.remove(compressed_file)


def cleanup_temp_dirs(target_path):
    if path.exists(target_path):
        shutil.rmtree(target_path)


def get_local_arch_download_path():
    current_os = constants.PLATFORM_TO_DOWNLOAD_PATH[os_util.get_os()]
    return f"{current_os}/{_get_local_arch_path(current_os)}"


def _get_local_arch_path(current_os):
    is_x64 = sys.maxsize > 2**32
    arch = "x64" if is_x64 else "x86"
    if current_os == constants.PLATFORM_TO_DOWNLOAD_PATH[commons_enums.PlatformsName.LINUX] and arch == "x64":
        return "x86_64"
    return arch


async def _download_tentacles(target_file, download_URL, aiohttp_session):
    async with aiohttp_session.get(download_URL) as resp:
        async with aiofiles.open(target_file, 'wb+') as downloaded_file:
            if resp.status != 200:
                raise RuntimeError(f"Can't find tentacle archive at: {download_URL} (status: {resp.status})")
            while True:
                chunk = await resp.content.read(DOWNLOADED_DATA_CHUNK_SIZE)
                if not chunk:
                    # resp.content.read returns an empty chunk when completed
                    break
                await downloaded_file.write(chunk)


async def _extract_tentacles(source_path, target_path, merge_dirs):
    if path.exists(target_path) and path.isdir(target_path) and not merge_dirs:
        shutil.rmtree(target_path)
    with zipfile.ZipFile(source_path) as zipped_tentacles:
        for archive_member in zipped_tentacles.namelist():
            if _is_tentacle_valid_tentacle_file(archive_member):
                zipped_tentacles.extract(archive_member, target_path)


def _is_tentacle_valid_tentacle_file(archive_member):
    member_path = archive_member.split("/")
    return len(member_path) >= 2 \
        and member_path[0] == constants.TENTACLES_ARCHIVE_ROOT \
        and (member_path[1] in constants.TENTACLE_TYPES or member_path[1] == constants.TENTACLES_PACKAGE_PROFILES_PATH)


def _is_url(string):
    return string.startswith("https://") \
           or string.startswith("http://") \
           or string.startswith("ftp://") \
           or string.startswith("sftp://")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import os
import shutil
import yaml

import octobot_tentacles_manager.exporters.artifact_exporter as artifact_exporter
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.util as util
import octobot_tentacles_manager.constants as constants


class TentacleBundleExporter(artifact_exporter.ArtifactExporter):
    def __init__(self,
                 artifact: models.TentaclePackage,
                 tentacles_folder: str,
                 output_dir: str = constants.DEFAULT_EXPORT_DIR,
                 should_zip: bool = False,
                 should_remove_artifacts_after_use: bool = False):
        super().__init__(artifact,
                         tentacles_folder=tentacles_folder,
                         output_dir=output_dir,
                         should_cythonize=False,
                         should_zip=should_zip,
                         with_dev_mode=False)
        self.should_remove_artifacts_after_use: bool = should_remove_artifacts_after_use

    async def prepare_export(self):
        if not os.path.exists(self.working_folder):
            os.makedirs(self.working_folder)

        for artifact in self.artifact.artifacts:
            # copy artifact content
            if os.path.isfile(artifact.output_path):
                shutil.copyfile(artifact.output_path, os.path.join(self.working_folder,
                                                                   os.path.basename(artifact.output_path)))
            else:
                if self.should_zip:
                    self.copy_directory_content_to_temporary_dir(artifact.output_path)
                else:
                    self.copy_directory_content_to_working_dir(artifact.output_path)
        self.create_metadata_file()

    def create_metadata_file(self) -> None:
        """
        Creates metadata file from artifacts count
        :return: None
        """
        if len(self.artifact.artifacts) == 1:
            artifact_metadata = models.MetadataFactory(self.artifact.artifacts[0]).create_metadata_instance()
        else:
            artifact_metadata = models.MetadataFactory(self.artifact).create_metadata_instance()
        with open(os.path.join(self.working_folder, constants.ARTIFACT_METADATA_FILE), "w") as metadata_file:
            metadata_file.write(yaml.dump(artifact_metadata.to_dict()))

    async def after_export(self) -> None:
        """
        Remove artifacts origin file or folder after bundling if necessary
        :return: None
        """
        if self.should_remove_artifacts_after_use:
            for artifact in self.artifact.artifacts:
                util.remove_dir_or_file_from_path(artifact.output_path)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.loaders import tentacle_loading

from octobot_tentacles_manager.loaders.tentacle_loading import (
    reload_tentacle_by_tentacle_class,
    get_tentacle_classes,
    ensure_tentacles_metadata,
    get_resources_path,
    get_tentacle_module_path,
    get_documentation_file_path,
    get_tentacle,
)

__all__ = [
    "reload_tentacle_by_tentacle_class",
    "get_tentacle_classes",
    "ensure_tentacles_metadata",
    "get_resources_path",
    "get_tentacle_module_path",
    "get_documentation_file_path",
    "get_tentacle",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import subprocess
import sys
import os
import os.path as path
import shutil

import octobot_commons.logging as logging

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.util as util


async def cythonize_and_compile_tentacles(directory):
    for element in os.scandir(directory):
        if element.name == constants.TENTACLE_METADATA:
            # this folder is a tentacle: cythonize
            # remove test folder
            test_folder = path.join(directory, constants.TENTACLE_TESTS)
            if path.exists(test_folder):
                shutil.rmtree(test_folder)
            await _cythonize_tentacle(directory)
            _compile_tentacle(directory)
            _clean_up_compiled_tentacle(directory)
            break
        elif element.is_dir():
            await cythonize_and_compile_tentacles(element)


async def _cythonize_tentacle(directory):
    # add missing .pxd files in directory and sub-directories
    # do not compile test files
    await _ensure_cython_header_files(directory)
    # update files
    if os.getenv('PACKAGER_UPDATE_FILES'):
        _update_files(directory)
    # add missing setup.py
    if not path.exists(path.join(directory, constants.SETUP_FILE)):
        await _create_setup_file(directory)


async def _ensure_cython_header_files(directory):
    files = []
    directories = []
    for element in os.scandir(directory):
        if element.is_file():
            files.append(element.name)
        elif element.is_dir() and element.name:
            directories.append(element)
    coros = []
    for element in files:
        element_simple_name = element.split(".")[0]
        header_file = f"{element_simple_name}{constants.CYTHON_EXT}"
        if element.endswith(constants.PYTHON_EXT) and header_file not in files:
            # this python file has no existing header file: create default one
            coros.append(util.find_or_create(path.join(directory, header_file),
                                             is_directory=False, file_content=constants.CYTHON_PXD_HEADER))
    for directory in directories:
        coros.append(_ensure_cython_header_files(directory))
    await asyncio.gather(*coros)


async def _create_setup_file(directory):
    package_list = _find_packages(directory)
    await util.find_or_create(path.join(directory, constants.SETUP_FILE), is_directory=False,
                              file_content=_get_setup_file_content(package_list))


def _find_packages(directory, parent_path=""):
    package_list = []
    for element in os.scandir(directory):
        if element.name.endswith(constants.PYTHON_EXT) and element.name != constants.PYTHON_INIT_FILE:
            package_list.append(f"{parent_path}{element.name.split(constants.PYTHON_EXT)[0]}")
        elif element.is_dir():
            package_list += _find_packages(element, parent_path=f"{parent_path}{element.name}.")
    return package_list


def _update_files(directory):
    # for later uses
    pass


def _compile_tentacle(directory):
    previous_dir = os.getcwd()
    os.chdir(directory)
    # Use subprocess.call() instead of sandbox.run_setup('setup.py', ['build_ext', '-i'])
    # to avoid multiple subsequent cythonization side effects.
    if subprocess.call([sys.executable, 'setup.py', 'build_ext', '-i']) != 0:
        logging.get_logger("CompiledPackageManager").error(
            f"Error when cythonizing {directory.path}, see above for details.")
    os.chdir(previous_dir)


def _clean_up_compiled_tentacle(directory):
    for element in os.scandir(directory):
        element_ext = element.name.split(".")[-1]
        if element.name in constants.COMPILED_TENTACLES_TO_REMOVE_FOLDERS or \
                (f".{element_ext}" in constants.COMPILED_TENTACLES_TO_REMOVE_ELEMENTS
                 and element.name not in constants.COMPILED_TENTACLES_TO_KEEP_ELEMENTS
                 and element.is_file()):
            if element.is_dir():
                shutil.rmtree(element)
            elif element.is_file():
                os.remove(element)
        elif element.is_dir():
            _clean_up_compiled_tentacle(element)


def _get_setup_file_content(package_list):
    return f"""
import os

from setuptools import dist

dist.Distribution().fetch_build_eggs(['Cython'])

try:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize
except ImportError:
    # create closure for deferred import
    def cythonize(*args, **kwargs):
        from Cython.Build import cythonize
        return cythonize(*args, **kwargs)

    def build_ext(*args, **kwargs):
        from Cython.Distutils import build_ext
        return build_ext(*args, **kwargs)

from setuptools import setup, Extension

packages_list = {package_list}

ext_modules = [
    Extension(package, [f"{{package.replace('.', '/')}}.py"])
    for package in packages_list]

setup(
    ext_modules=cythonize(ext_modules),
)
"""



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import aiohttp

import octobot_tentacles_manager.uploaders.uploader as uploader


class NexusUploader(uploader.Uploader):
    ENV_NEXUS_USERNAME = "NEXUS_USERNAME"
    ENV_NEXUS_PASSWORD = "NEXUS_PASSWORD"
    ENV_NEXUS_URL = "NEXUS_URL"
    NEXUS_EXPECTED_RESPONSE_STATUS = [200, 201]

    def __init__(self):
        super().__init__()
        self.aiohttp_session: aiohttp.ClientSession = None
        self.nexus_username: str = os.getenv(NexusUploader.ENV_NEXUS_USERNAME, None)
        self.nexus_password: str = os.getenv(NexusUploader.ENV_NEXUS_PASSWORD, None)
        self.nexus_url: str = os.getenv(NexusUploader.ENV_NEXUS_URL, None)
        if None in [self.nexus_username, self.nexus_password, self.nexus_url]:
            raise TypeError("Some nexus environment variables are missing, "
                            "please ensure that NEXUS_USERNAME, NEXUS_PASSWORD and NEXUS_URL are defined.")

    async def upload_file(self, upload_path: str, file_path: str, destination_file_name: str = None) -> int:
        """
        Upload file on nexus wrapper
        :param upload_path: the upload path, the internal path after self.nexus_url
        :param file_path: the file local path
        :param destination_file_name: the file name on nexus (optional : default file_path basename)
        :return: the result of _upload
        """
        dest_file_name: str = destination_file_name if destination_file_name is not None else os.path.basename(
            file_path)
        upload_file_url: str = f"{self.nexus_url}/{upload_path}/{dest_file_name}"
        self.logger.info(f"Uploading {file_path} to nexus at {upload_file_url}...")
        return await self._upload(file_url_on_nexus=upload_file_url, local_file_path=file_path)

    async def upload_folder(self, upload_path: str, folder_path: str, destination_folder_name: str = None) -> int:
        """
        Upload folder content on nexus wrapper
        :param upload_path: the upload path, the internal path after self.nexus_url
        :param folder_path: the folder local path
        :param destination_folder_name: the folder name on nexus (optional : default folder_path basename)
        :return: the sum of all of _upload returns
        """
        error_count: int = 0
        dest_folder_name: str = destination_folder_name \
            if destination_folder_name is not None else os.path.basename(folder_path)
        upload_folder_url: str = f"{self.nexus_url}/{upload_path}/{dest_folder_name}"
        for file_path in os.listdir(folder_path):
            upload_file_url = f"{upload_folder_url}/{file_path}"
            self.logger.debug(f"Uploading {file_path} to nexus at {upload_file_url}...")
            error_count += await self._upload(file_url_on_nexus=upload_file_url,
                                              local_file_path=os.path.join(folder_path, file_path))
        return error_count

    async def create_session(self) -> None:
        """
        Create aiohttp session if necessary
        :return: None
        """
        if self.aiohttp_session is None:
            self.aiohttp_session = aiohttp.ClientSession()

    async def close_session(self) -> None:
        """
        Close aiohttp session if necessary
        :return: None
        """
        if self.aiohttp_session is not None:
            await self.aiohttp_session.close()

    async def _upload(self, file_url_on_nexus: str, local_file_path: str) -> int:
        """
        Upload a file on nexus
        :param file_url_on_nexus: the complete upload url
        :param local_file_path: the local file path
        :return: 0 if upload succeed else 1
        """
        await self.create_session()
        with open(local_file_path, 'rb') as file_content:
            response = await self.aiohttp_session.request('put',
                                                          file_url_on_nexus,
                                                          data=file_content,
                                                          auth=aiohttp.BasicAuth(self.nexus_username,
                                                                                 self.nexus_password))
            if response.status not in NexusUploader.NEXUS_EXPECTED_RESPONSE_STATUS:
                self.logger.error(f"Failed to upload file on nexus "
                                  f"(status code {response.status}) : {await response.text()}")
                return 1
        return 0



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_tentacles_manager.api as util
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.workers as workers

USER_HELP = """Update the given tentacle modules and install missing requirements if any.
Does not update already filled requirements regardless of their version to avoid conflicts.
Prefer a full tentacles update if a requirement version is creating conflicts.
    Does not edit tentacles configuration files."""


async def update_all_tentacles(tentacles_path_or_url, tentacle_path=constants.TENTACLES_PATH,
                               bot_path=constants.DEFAULT_BOT_PATH, aiohttp_session=None, quite_mode=False) -> int:
    return await _update_tentacles(None, tentacles_path_or_url, tentacle_path,
                                   bot_path, aiohttp_session=aiohttp_session, quite_mode=quite_mode)


async def update_tentacles(tentacle_names, tentacles_path_or_url, tentacle_path=constants.TENTACLES_PATH,
                           bot_path=constants.DEFAULT_BOT_PATH, aiohttp_session=None, quite_mode=False) -> int:
    return await _update_tentacles(tentacle_names, tentacles_path_or_url,
                                   tentacle_path, bot_path, aiohttp_session=aiohttp_session, quite_mode=quite_mode)


async def _update_tentacles(tentacle_names, tentacles_path_or_url, tentacle_path=constants.TENTACLES_PATH,
                            bot_path=constants.DEFAULT_BOT_PATH, use_confirm_prompt=False, aiohttp_session=None,
                            quite_mode=False) -> int:
    update_worker = workers.UpdateWorker(constants.TENTACLES_INSTALL_TEMP_DIR, tentacle_path, bot_path,
                                         use_confirm_prompt, aiohttp_session, quite_mode=quite_mode)
    return await util.manage_tentacles(update_worker, tentacle_names, tentacles_path_or_url, aiohttp_session)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path

import octobot_commons.tentacles_management as tentacles_management

import octobot_tentacles_manager.api as api
import octobot_tentacles_manager.configuration as configuration
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.util as util


async def ensure_setup_configuration(tentacle_path=constants.TENTACLES_PATH, bot_path=constants.DEFAULT_BOT_PATH,
                                     bot_install_dir=constants.DEFAULT_BOT_INSTALL_DIR) -> None:
    if not path.exists(path.join(bot_path, constants.USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH)):
        await api.repair_installation(tentacle_path, bot_path, bot_install_dir, verbose=False)


def get_tentacles_setup_config(
        config_path=constants.USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH) -> configuration.TentaclesSetupConfiguration:
    setup_config = configuration.TentaclesSetupConfiguration(config_path=config_path)
    setup_config.read_config()
    return setup_config


def create_tentacles_setup_config_with_tentacles(*tentacles_classes):
    setup_config = configuration.TentaclesSetupConfiguration()
    setup_config.from_activated_tentacles_classes(*tentacles_classes)
    return setup_config


def is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config: configuration.TentaclesSetupConfiguration,
                                                    klass_name: str, default_value=False, raise_errors=False) -> bool:
    try:
        return tentacles_setup_config.is_tentacle_activated(klass_name)
    except KeyError as e:
        if raise_errors:
            raise e
        return default_value


def get_class_from_name_with_activated_required_tentacles(name: str,
                                                          parent_class,
                                                          tentacles_setup_config: configuration.TentaclesSetupConfiguration,
                                                          with_class_method=None):
    for subclass in tentacles_management.get_all_classes_from_parent(parent_class):
        # Filter sub classes to only use the one that is appropriate to the given name and that has its
        # tentacles requirements activated (identified by REQUIRED_ACTIVATED_TENTACLES iterable class attribute).
        if subclass.get_name() == name and \
            tentacles_setup_config is not None and \
            all(is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config, required_tentacle.__name__)
                for required_tentacle in subclass.REQUIRED_ACTIVATED_TENTACLES):
            if with_class_method is None or getattr(subclass, with_class_method)():
                return subclass
    return None


def get_tentacles_activation(tentacles_setup_config: configuration.TentaclesSetupConfiguration) -> dict:
    return tentacles_setup_config.tentacles_activation


def get_tentacles_config_folder(tentacles_setup_config: configuration.TentaclesSetupConfiguration) -> str:
    return tentacles_setup_config.get_config_folder()


def get_registered_tentacle_packages(tentacles_setup_config: configuration.TentaclesSetupConfiguration) -> dict:
    return tentacles_setup_config.registered_tentacles


def unregister_tentacle_packages(tentacles_setup_config: configuration.TentaclesSetupConfiguration,
                                 package_name: str) -> dict:
    return tentacles_setup_config.unregister_tentacles_package(package_name)


def update_activation_configuration(tentacles_setup_config: configuration.TentaclesSetupConfiguration,
                                    new_config: dict, deactivate_others: bool, add_missing_elements=False) \
        -> bool:
    return tentacles_setup_config.update_activation_configuration(new_config, deactivate_others, add_missing_elements)


def save_tentacles_setup_configuration(tentacles_setup_config: configuration.TentaclesSetupConfiguration) -> None:
    tentacles_setup_config.save_config()


def get_activated_tentacles(tentacles_setup_config: configuration.TentaclesSetupConfiguration) -> list:
    return [tentacle_class
            for tentacle_classes in tentacles_setup_config.tentacles_activation.values()
            for tentacle_class, activated in tentacle_classes.items()
            if activated]


def update_tentacle_config(tentacles_setup_config: configuration.TentaclesSetupConfiguration,
                           tentacle_class: object, config_data: dict) -> None:
    configuration.update_config(tentacles_setup_config, tentacle_class, config_data)


def get_tentacle_config(tentacles_setup_config: configuration.TentaclesSetupConfiguration, klass) -> dict:
    return configuration.get_config(tentacles_setup_config, klass)


def factory_tentacle_reset_config(tentacles_setup_config: configuration.TentaclesSetupConfiguration, klass) -> None:
    return configuration.factory_reset_config(tentacles_setup_config, klass)


def get_tentacle_config_schema_path(klass) -> str:
    return configuration.get_config_schema_path(klass)


def get_compiled_tentacles_url(base_url, version) -> str:
    return f"{base_url}{util.get_local_arch_download_path()}/{version}.zip"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import shutil

import pytest
import yaml

import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.exporters as exporters
import octobot_tentacles_manager.util as util
import octobot_tentacles_manager.constants as constants
from tests.api import install_tentacles

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_tentacle_bundle_exporter_for_each_tentacle(install_tentacles):
    # Export each tentacle in a bundle
    for tentacle in util.load_tentacle_with_metadata(constants.TENTACLES_PATH):
        tentacle_package = models.TentaclePackage()
        await exporters.TentacleExporter(artifact=tentacle, should_zip=True,
                                         tentacles_folder=constants.TENTACLES_PATH).export()
        tentacle_package.add_artifact(tentacle)
        await exporters.TentacleBundleExporter(
            artifact=tentacle_package,
            tentacles_folder=constants.TENTACLES_PATH).export()

    # Check if each tentacle bundle has been generated
    # check files count
    output_files = os.listdir(constants.DEFAULT_EXPORT_DIR)
    assert len(output_files) == 20
    assert "daily_trading_mode.zip" in output_files
    assert "generic_exchange_importer@1.2.0" in output_files
    assert "other_instant_fluctuations_evaluator@1.2.0" in output_files
    assert "mixed_strategies_evaluator.zip" in output_files
    assert "mixed_strategies_evaluator" not in output_files


async def test_tentacle_bundle_exporter_for_an_unique_bundle_containing_all_tentacles(install_tentacles):
    # Export all tentacles and generate a bundle containing all
    tentacle_package = models.TentaclePackage()
    for tentacle in util.load_tentacle_with_metadata(constants.TENTACLES_PATH):
        await exporters.TentacleExporter(artifact=tentacle, should_zip=True,
                                         tentacles_folder=constants.TENTACLES_PATH).export()
        tentacle_package.add_artifact(tentacle)
    await exporters.TentacleBundleExporter(
        artifact=tentacle_package,
        tentacles_folder=constants.TENTACLES_PATH,
        should_remove_artifacts_after_use=True).export()

    # Check if the final bundle contains all exported tentacles and a metadata file
    # check files count
    output_files = os.listdir(constants.DEFAULT_EXPORT_DIR)
    assert len(output_files) == 1
    exported_bundle_path = os.path.join(constants.DEFAULT_EXPORT_DIR, output_files[0])
    output_files = os.listdir(exported_bundle_path)
    assert len(output_files) == 11
    assert "daily_trading_mode.zip" in output_files
    assert "generic_exchange_importer@1.2.0_package" not in output_files
    assert "other_instant_fluctuations_evaluator@1.2.0_package" not in output_files
    assert "mixed_strategies_evaluator.zip" in output_files
    assert "mixed_strategies_evaluator" not in output_files
    assert constants.ARTIFACT_METADATA_FILE in output_files

    # test multiple tentacle bundle metadata
    with open(os.path.join(exported_bundle_path, constants.ARTIFACT_METADATA_FILE)) as metadata_file:
        metadata_content = yaml.load(metadata_file.read(), Loader=yaml.SafeLoader)
        assert metadata_content[constants.ARTIFACT_METADATA_ARTIFACT_TYPE] == "tentacle_package"
        assert len(metadata_content[constants.ARTIFACT_METADATA_ARTIFACTS]) == 10
        assert "forum_evaluator" in metadata_content[constants.ARTIFACT_METADATA_ARTIFACTS]


async def test_tentacle_bundle_exporter_with_specified_output_dir(install_tentacles):
    specified_output_dir = "out/dir/test"
    # Export each tentacle in a bundle in a specified output dir
    for tentacle in util.load_tentacle_with_metadata(constants.TENTACLES_PATH):
        tentacle_package = models.TentaclePackage()
        await exporters.TentacleExporter(artifact=tentacle,
                                         should_zip=True,
                                         output_dir=specified_output_dir,
                                         tentacles_folder=constants.TENTACLES_PATH).export()
        tentacle_package.add_artifact(tentacle)
        await exporters.TentacleBundleExporter(
            artifact=tentacle_package,
            output_dir=specified_output_dir,
            tentacles_folder=constants.TENTACLES_PATH).export()

    # Check if each tentacle bundle has been generated in the specified directory
    output_files = os.listdir(specified_output_dir)
    assert len(output_files) == 20
    assert "daily_trading_mode.zip" in output_files
    assert "generic_exchange_importer@1.2.0" in output_files
    shutil.rmtree(specified_output_dir)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.exporters import artifact_exporter
from octobot_tentacles_manager.exporters.artifact_exporter import (
    ArtifactExporter,
)

from octobot_tentacles_manager.exporters import tentacle_exporter
from octobot_tentacles_manager.exporters import tentacle_package_exporter
from octobot_tentacles_manager.exporters import tentacle_bundle_exporter

from octobot_tentacles_manager.exporters.tentacle_exporter import (
    TentacleExporter,
)
from octobot_tentacles_manager.exporters.tentacle_package_exporter import (
    TentaclePackageExporter,
)
from octobot_tentacles_manager.exporters.tentacle_bundle_exporter import (
    TentacleBundleExporter,
)

__all__ = [
    "ArtifactExporter",
    "TentacleExporter",
    "TentaclePackageExporter",
    "TentacleBundleExporter",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiofiles
import os
import os.path as path
import shutil

import octobot_tentacles_manager.constants as constants


async def find_or_create(path_to_create, is_directory=True, file_content=""):
    if not path.exists(path_to_create):
        if is_directory:
            if not path.isdir(path_to_create):
                os.makedirs(path_to_create)
        else:
            if not path.isfile(path_to_create):
                # should be used for python init.py files only
                async with aiofiles.open(path_to_create, "w+") as file:
                    await file.write(file_content)
        return True
    return False


async def replace_with_remove_or_rename(new_file_or_dir_entry, dest_file_or_dir):
    try:
        if path.isfile(dest_file_or_dir):
            os.remove(dest_file_or_dir)
        else:
            if path.exists(dest_file_or_dir):
                shutil.rmtree(dest_file_or_dir)
    except PermissionError:
        # can't remove file / folder: might be locked (ex: .pyd files)
        # move it into TO_REMOVE_FOLDER for it to be deleted afterwards
        dest_path, dest_file = path.split(dest_file_or_dir)
        to_rm_folder = path.join(dest_path, constants.TO_REMOVE_FOLDER)
        await find_or_create(to_rm_folder, is_directory=True)
        shutil.move(dest_file_or_dir, path.join(to_rm_folder, dest_file))
    if path.isfile(new_file_or_dir_entry):
        shutil.copyfile(new_file_or_dir_entry, dest_file_or_dir)
    else:
        shutil.copytree(new_file_or_dir_entry, dest_file_or_dir)


def merge_folders(to_merge_folder, dest_folder, ignore_func=None):
    dest_folder_elements = {
        element.name: element for element in os.scandir(dest_folder)
    }
    elements = list(os.scandir(to_merge_folder))
    ignored_elements = ignore_func(to_merge_folder, (e.name for e in elements)) if ignore_func is not None else []
    filtered_elements = [element
                         for element in elements
                         if element.name not in ignored_elements]
    for element in filtered_elements:
        dest = path.join(dest_folder, element.name)
        if element.is_file():
            shutil.copy(element.path, dest)
        else:
            if element.name not in dest_folder_elements:
                shutil.copytree(element.path, dest, ignore=ignore_func if ignore_func is not None else None)
            else:
                merge_folders(element.path, dest_folder_elements[element.name].path, ignore_func=ignore_func)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.api as util
import octobot_tentacles_manager.workers as workers


USER_HELP = """Uninstall the given tentacle modules.
    Does not delete the associated tentacle configuration."""


async def uninstall_all_tentacles(tentacle_path=constants.TENTACLES_PATH,
                                  bot_path=constants.DEFAULT_BOT_PATH,
                                  use_confirm_prompt=False,
                                  quite_mode=False,
                                  setup_config=None) -> int:
    return await _uninstall_tentacles(None, tentacle_path,
                                      bot_path,
                                      use_confirm_prompt,
                                      quite_mode=quite_mode,
                                      tentacles_setup_config_to_update=setup_config)


async def uninstall_tentacles(tentacle_names,
                              tentacle_path=constants.TENTACLES_PATH,
                              bot_path=constants.DEFAULT_BOT_PATH,
                              use_confirm_prompt=False,
                              quite_mode=False) -> int:
    return await _uninstall_tentacles(tentacle_names,
                                      tentacle_path,
                                      bot_path,
                                      use_confirm_prompt,
                                      quite_mode=quite_mode)


async def _uninstall_tentacles(tentacle_names,
                               tentacle_path=constants.TENTACLES_PATH,
                               bot_path=constants.DEFAULT_BOT_PATH,
                               use_confirm_prompt=False,
                               quite_mode=False,
                               tentacles_setup_config_to_update=None) -> int:
    uninstall_worker = workers.UninstallWorker(None, tentacle_path, bot_path, use_confirm_prompt,
                                               None, quite_mode=quite_mode)
    uninstall_worker.tentacles_setup_config_to_update = tentacles_setup_config_to_update
    return await util.manage_tentacles(uninstall_worker, tentacle_names)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import shutil
import aiohttp
import pytest

import octobot_tentacles_manager.api as api
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.managers as managers

TENTACLE_PACKAGE = "tentacle_package"
TEST_EXPORT_DIR = "test_export_dir"


@pytest.yield_fixture()
async def install_tentacles():
    _cleanup()
    async with aiohttp.ClientSession() as session:
        assert await api.install_all_tentacles(os.path.join("tests", "static", "tentacles.zip"),
                                               aiohttp_session=session) == 0
        yield
    _cleanup()


def _cleanup():
    if os.path.exists(constants.TENTACLES_PATH):
        managers.TentaclesSetupManager.delete_tentacles_arch(force=True)
    if os.path.exists(constants.TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER):
        shutil.rmtree(constants.TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
    if os.path.exists(TENTACLE_PACKAGE):
        shutil.rmtree(TENTACLE_PACKAGE)
    if os.path.exists(TEST_EXPORT_DIR):
        shutil.rmtree(TEST_EXPORT_DIR)
    if os.path.exists(constants.DEFAULT_EXPORT_DIR):
        shutil.rmtree(constants.DEFAULT_EXPORT_DIR)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import os.path as path

import octobot_tentacles_manager.exporters.artifact_exporter as artifact_exporter
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.util as util


class TentaclePackageExporter(artifact_exporter.ArtifactExporter):
    def __init__(self,
                 artifact: models.TentaclePackage,
                 exported_tentacles_package: str,
                 tentacles_folder: str,
                 output_dir: str = constants.DEFAULT_EXPORT_DIR,
                 should_cythonize: bool = False,
                 should_zip: bool = False,
                 with_dev_mode: bool = False):
        super().__init__(artifact,
                         tentacles_folder=tentacles_folder,
                         output_dir=output_dir,
                         should_cythonize=should_cythonize,
                         should_zip=should_zip,
                         with_dev_mode=with_dev_mode)
        self.exported_tentacles_package: str = exported_tentacles_package
        self.should_cleanup_working_folder: bool = True

        self.tentacles_filter: util.TentacleFilter = None
        self.tentacles_white_list: list = []
        self.tentacles = []

    async def prepare_export(self):
        if not self.with_dev_mode or self.exported_tentacles_package is not None:
            self.tentacles = util.load_tentacle_with_metadata(self.tentacles_folder)
            # remove dev-mode or non exported package tentacles if necessary
            self.tentacles_white_list = util.filter_tentacles_by_dev_mode_and_package(
                tentacles=self.tentacles,
                with_dev_mode=self.with_dev_mode,
                package_filter=self.exported_tentacles_package
            )

        self.tentacles_filter = util.TentacleFilter(self.tentacles, self.tentacles_white_list)
        if self.should_zip:
            self.copy_directory_content_to_temporary_dir(self.tentacles_folder,
                                                         ignore=self.tentacles_filter.should_ignore)
        else:
            self.copy_directory_content_to_working_dir(self.tentacles_folder,
                                                       ignore=self.tentacles_filter.should_ignore)

    async def after_export(self):
        pass



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import octobot_commons.profiles as profiles
import octobot_tentacles_manager.constants as constants


def get_profile_folders(package_root) -> list:
    profiles_source_path = os.path.join(package_root, constants.TENTACLES_PACKAGE_PROFILES_PATH)
    if os.path.isdir(profiles_source_path):
        return [
            profile.path
            for profile in os.scandir(profiles_source_path)
        ]
    return []


def import_profile(profile_path, bot_install_root_folder) -> None:
    profile_name = os.path.split(profile_path)[-1]
    profiles.import_profile(profile_path, name=profile_name, bot_install_path=bot_install_root_folder)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.api import configurator
from octobot_tentacles_manager.api import updater
from octobot_tentacles_manager.api import loader
from octobot_tentacles_manager.api import inspector
from octobot_tentacles_manager.api import uninstaller
from octobot_tentacles_manager.api import creator
from octobot_tentacles_manager.api import installer
from octobot_tentacles_manager.api import util
from octobot_tentacles_manager.api import uploader

from octobot_tentacles_manager.api.util import (
    manage_tentacles,
)

from octobot_tentacles_manager.api.uploader import (
    upload_file_to_nexus,
    upload_folder_to_nexus,
    upload_file_or_folder_to_nexus,
)

from octobot_tentacles_manager.api.configurator import (
    get_tentacles_setup_config,
    create_tentacles_setup_config_with_tentacles,
    is_tentacle_activated_in_tentacles_setup_config,
    get_class_from_name_with_activated_required_tentacles,
    get_tentacles_activation,
    get_registered_tentacle_packages,
    unregister_tentacle_packages,
    update_activation_configuration,
    save_tentacles_setup_configuration,
    get_activated_tentacles,
    update_tentacle_config,
    get_tentacle_config,
    factory_tentacle_reset_config,
    get_tentacle_config_schema_path,
    get_compiled_tentacles_url,
    ensure_setup_configuration,
)
from octobot_tentacles_manager.api.updater import (
    update_all_tentacles,
    update_tentacles,
)
from octobot_tentacles_manager.api.loader import (
    load_tentacles,
    reload_tentacle_info,
)
from octobot_tentacles_manager.api.inspector import (
    get_installed_tentacles_modules,
    get_tentacle_group,
    get_tentacle_version,
    get_tentacle_origin_package,
    get_tentacle_module_name,
    get_tentacle_resources_path,
    get_tentacle_documentation_path,
    check_tentacle_version,
)
from octobot_tentacles_manager.api.uninstaller import (
    uninstall_all_tentacles,
    uninstall_tentacles,
)
from octobot_tentacles_manager.api.creator import (
    start_tentacle_creator,
    create_tentacles_package,
    create_all_tentacles_bundle,
)
from octobot_tentacles_manager.api.installer import (
    install_all_tentacles,
    install_tentacles,
    install_single_tentacle,
    repair_installation,
)

__all__ = [
    "get_tentacles_setup_config",
    "create_tentacles_setup_config_with_tentacles",
    "is_tentacle_activated_in_tentacles_setup_config",
    "get_class_from_name_with_activated_required_tentacles",
    "get_tentacles_activation",
    "get_registered_tentacle_packages",
    "unregister_tentacle_packages",
    "update_activation_configuration",
    "save_tentacles_setup_configuration",
    "get_activated_tentacles",
    "update_tentacle_config",
    "get_tentacle_config",
    "factory_tentacle_reset_config",
    "get_tentacle_config_schema_path",
    "get_compiled_tentacles_url",
    "ensure_setup_configuration",
    "update_all_tentacles",
    "update_tentacles",
    "load_tentacles",
    "reload_tentacle_info",
    "get_installed_tentacles_modules",
    "get_tentacle_group",
    "get_tentacle_version",
    "get_tentacle_origin_package",
    "get_tentacle_module_name",
    "get_tentacle_resources_path",
    "get_tentacle_documentation_path",
    "check_tentacle_version",
    "uninstall_all_tentacles",
    "uninstall_tentacles",
    "start_tentacle_creator",
    "create_tentacles_package",
    "install_all_tentacles",
    "install_tentacles",
    "install_single_tentacle",
    "repair_installation",
    "manage_tentacles",
    "create_all_tentacles_bundle",
    "upload_file_to_nexus",
    "upload_folder_to_nexus",
    "upload_file_or_folder_to_nexus",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import os.path as path
import shutil

import octobot_commons.logging as logging

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.managers as managers
import octobot_tentacles_manager.util as util


class TentacleManager:

    def __init__(self, tentacle, bot_installation_path=constants.DEFAULT_BOT_PATH):
        self.tentacle = tentacle
        self.bot_installation_path = bot_installation_path
        self.target_tentacle_path = None

    async def install_tentacle(self, tentacle_path):
        self.target_tentacle_path = path.join(tentacle_path, self.tentacle.tentacle_type.to_path())
        tentacle_module_path = path.join(self.target_tentacle_path, self.tentacle.name)
        await self._update_tentacle_folder(tentacle_module_path)
        await managers.create_tentacle_init_file_if_necessary(tentacle_module_path, self.tentacle)
        self.import_tentacle_config_if_any(tentacle_module_path)

    async def uninstall_tentacle(self):
        shutil.rmtree(path.join(self.bot_installation_path, self.tentacle.tentacle_path, self.tentacle.name))

    @staticmethod
    def find_tentacles_missing_requirements(tentacle, to_install_version_by_modules, available_tentacles):
        # check if requirement is in tentacles to be installed in this call
        return {
            requirement: version
            for requirement, version in tentacle.extract_tentacle_requirements()
            if not TentacleManager.is_requirement_satisfied(requirement, version, tentacle,
                                                            to_install_version_by_modules, available_tentacles)
        }

    @staticmethod
    def is_requirement_satisfied(requirement, version, tentacle, to_install_version_by_modules, available_tentacles):
        satisfied = False
        # check in to install tentacles
        if requirement in to_install_version_by_modules:
            satisfied = TentacleManager._ensure_version(tentacle.name,
                                                        version,
                                                        to_install_version_by_modules[requirement])
        if not satisfied:
            # check in available tentacles
            for available_tentacle in available_tentacles:
                if available_tentacle.name == requirement:
                    return TentacleManager._ensure_version(tentacle.name,
                                                           version,
                                                           available_tentacle.version)
        return satisfied

    @staticmethod
    def _ensure_version(name, version, available_version):
        if version is None:
            return True
        elif version != available_version:
            logging.get_logger(TentacleManager.__name__). \
                error(f"Incompatible tentacle version requirement for "
                      f"{name}: requires {version}, installed: "
                      f"{available_version}. This tentacle might not work as expected")
            return True
        return False

    async def _update_tentacle_folder(self, target_tentacle_path):
        reference_tentacle_path = path.join(self.tentacle.tentacle_path, self.tentacle.name)
        await util.find_or_create(target_tentacle_path)
        for tentacle_file_entry in os.scandir(reference_tentacle_path):
            await util.replace_with_remove_or_rename(tentacle_file_entry,
                                                     path.join(target_tentacle_path, tentacle_file_entry.name))

    def import_tentacle_config_if_any(self, tentacle_module_path, replace=False):
        target_tentacle_config_path = path.join(tentacle_module_path, constants.TENTACLE_CONFIG)
        if path.isdir(target_tentacle_config_path):
            for config_file_entry in os.scandir(target_tentacle_config_path):
                if config_file_entry.name.endswith(constants.CONFIG_EXT) \
                        and not config_file_entry.name.endswith(constants.CONFIG_SCHEMA_EXT):
                    target_user_path = \
                        path.join(self.bot_installation_path,
                                  constants.USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH,
                                  config_file_entry.name)
                    if replace or not path.exists(target_user_path):
                        shutil.copyfile(path.join(target_tentacle_config_path, config_file_entry.name),
                                        target_user_path)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.loaders as loaders
import octobot_tentacles_manager.util as util


async def manage_tentacles(worker, tentacle_names, tentacles_path_or_url=None, aiohttp_session=None) -> int:
    errors_count = 0
    logger = logging.get_logger(__name__)
    try:
        if tentacles_path_or_url is not None:
            await util.fetch_and_extract_tentacles(constants.TENTACLES_INSTALL_TEMP_DIR,
                                                   tentacles_path_or_url,
                                                   aiohttp_session)
        errors_count = await worker.process(tentacle_names)
    except Exception as e:
        logger.exception(e, True, f"Exception during {worker.__class__.__name__} processing: {e}")
        # ensure error is taken into account
        if errors_count == 0:
            errors_count = 1
    finally:
        if tentacles_path_or_url is not None:
            util.cleanup_temp_dirs(constants.TENTACLES_INSTALL_TEMP_DIR)
        try:
            # reload tentacles data
            loaders.reload_tentacle_by_tentacle_class()
        except Exception as e:
            logger.exception(e, True, f"Exception while reloading tentacles data: {e}")
            errors_count = 1
    return errors_count



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.models import artifact
from octobot_tentacles_manager.models.artifact import (
    Artifact,
)

from octobot_tentacles_manager.models import tentacle_type
from octobot_tentacles_manager.models import tentacle
from octobot_tentacles_manager.models import tentacle_package
from octobot_tentacles_manager.models import tentacle_package

from octobot_tentacles_manager.models.tentacle_type import (
    TentacleType,
)
from octobot_tentacles_manager.models.tentacle import (
    Tentacle,
)
from octobot_tentacles_manager.models.profile import (
    Profile,
)
from octobot_tentacles_manager.models.tentacle_package import (
    TentaclePackage,
)

from octobot_tentacles_manager.models import tentacle_factory
from octobot_tentacles_manager.models.tentacle_factory import (
    TentacleFactory,
)

from octobot_tentacles_manager.models import metadata
from octobot_tentacles_manager.models.metadata import (
    ArtifactMetadata,
    TentacleMetadata,
    TentaclePackageMetadata,
    ProfileMetadata,
    MetadataFactory,
)

__all__ = [
    "Artifact",
    "TentacleType",
    "TentacleFactory",
    "Tentacle",
    "TentaclePackage",
    "Profile",
    "ArtifactMetadata",
    "TentacleMetadata",
    "TentaclePackageMetadata",
    "ProfileMetadata",
    "MetadataFactory",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import os.path as path
import sys

import octobot_commons.enums as enums
import octobot_commons.logging as logging
import octobot_commons.os_util as os_util


# TODO: adapt this wrapper to handle python modules requirements
# Warning: Impossible to dynamically load python modules with unmet dependencies in frozen python environment
# (ex: compiled binary). However these dynamic imports work with Python / Docker environment.
# code from pip POC (https://github.com/Drakkar-Software/OctoBot-Tentacles-Manager/tree/pip-usage-poc)

# TODO: remove from to .coveragerc when adapted

"""
Tentacles management
"""


async def install_tentacle(tentacle_path: str, tentacle_name: str) -> (str, str):
    return await _run_pip_install(path.join(tentacle_path, "Trading"), tentacle_name)


async def update_tentacle(tentacle_path: str, tentacle_name: str) -> (str, str):
    return await _run_pip_update(path.join(tentacle_path, "Trading"), tentacle_name)


async def list_installed_tentacles(tentacle_path: str) -> list:
    return [await _run_pip_freeze(path.join(tentacle_path, tentacle_type_path))
            for tentacle_type_path in ["Trading"]]

"""
Pip wrapper
"""


async def _run_pip_command(args) -> (str, str):
    if os_util.get_octobot_type() == enums.OctoBotTypes.BINARY.value:
        raise RuntimeError("Can't use PIP in a frozen binary environment")
    # Create subprocess
    process = await asyncio.create_subprocess_exec(
        sys.executable, "-m", "pip", *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    # Status
    logging.get_logger().info(f"Started: {args}, pid={process.pid}")

    # Wait for the subprocess to finish
    stdout, stderr = await process.communicate()

    # Progress
    if process.returncode == 0:
        logging.get_logger().info(f"Done: {args}, pid={process.pid}")
    else:
        logging.get_logger().error(f"Failed: {args}, pid={process.pid}, result: {_parse_pip_command_result(stderr)}")

    return stdout, stderr


"""
Pip freeze wrapper
"""


def _get_pip_freeze_args(path: str = None) -> list:
    args = ["freeze"]
    if path is not None:
        args.append("--path")
        args.append(path)
    return args


async def _run_pip_freeze(path: str = None) -> (str, str):
    return await _run_pip_command(_get_pip_freeze_args(path=path))


"""
Pip install wrapper
"""


def _get_pip_install_args(path: str = None, package_name: str = None) -> list:
    args = ["install"]
    if path is not None:
        args.append("-t")
        args.append(path)
    if package_name is not None:
        args.append(package_name)
    return args


async def _run_pip_install(path: str = None, package_name: str = None) -> (str, str):
    return await _run_pip_command(_get_pip_install_args(path=path, package_name=package_name))


"""
Pip update wrapper
"""


def _get_pip_update_args(path: str = None, package_name: str = None) -> list:
    args = _get_pip_install_args(path=path, package_name=package_name)
    args.append("-U")
    return args


async def _run_pip_update(path: str = None, package_name: str = None) -> (str, str):
    return await _run_pip_command(_get_pip_update_args(path=path, package_name=package_name))


"""
Pip utils
"""


def _parse_pip_command_result(stdout) -> str:
    return stdout.decode().strip()



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.models.metadata import artifact_metadata
from octobot_tentacles_manager.models.metadata.artifact_metadata import (
    ArtifactMetadata,
)

from octobot_tentacles_manager.models.metadata import tentacle_metadata
from octobot_tentacles_manager.models.metadata import tentacle_package_metadata
from octobot_tentacles_manager.models.metadata import profile_metadata

from octobot_tentacles_manager.models.metadata.tentacle_metadata import (
    TentacleMetadata,
)
from octobot_tentacles_manager.models.metadata.profile_metadata import (
    ProfileMetadata,
)
from octobot_tentacles_manager.models.metadata.tentacle_package_metadata import (
    TentaclePackageMetadata,
)

from octobot_tentacles_manager.models.metadata import metadata_factory
from octobot_tentacles_manager.models.metadata.metadata_factory import (
    MetadataFactory,
)

__all__ = [
    "ArtifactMetadata",
    "TentacleMetadata",
    "TentaclePackageMetadata",
    "ProfileMetadata",
    "MetadataFactory",
]
